function _asyncToGenerator(fn) { return function () { var gen = fn.apply(this, arguments); return new Promise(function (resolve, reject) { function step(key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { return Promise.resolve(value).then(function (value) { step("next", value); }, function (err) { step("throw", err); }); } } return step("next"); }); }; }

const { Writable } = require('stream');
const Box = require('./box');
const nodeify = require('nodeify');

const EOS = Symbol('EOS');

module.exports = class IteratorWritable extends Writable {

  constructor(opts) {
    var _this;

    _this = super(opts);
    this.writeAsync = _asyncToGenerator(function* () {
      let err = new Error("Must call run() first.");
      _this.emit('error', err);
      throw err;
    });
    this.endRequest = () => this.emit('error', new Error("Must call run() first."));
    this.waitFirstChunk = new Box();
    this.run();
  }

  run() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      try {
        // Make sure the stream has any item first.
        _this2.next = function () {
          throw new Error("Cannot request for an iteration yet.");
        };

        // Read first chunk.
        let chunk = yield _this2.readChunkFromStream();
        let waitNextIteration = new Box(); // dummy
        if (chunk !== EOS) {
          let previousChunk = chunk;
          // Prepare stream to wait for the next chunk.
          let streamWaitNext = new Box();

          // Loop condition: Tell if the previous signal was the end of the
          // stream + receive a chunk from the stream
          let readChunk = (() => {
            var _ref2 = _asyncToGenerator(function* () {
              if (previousChunk === EOS) {
                return false;
              }
              chunk = yield _this2.readChunkFromStream(streamWaitNext);
              return true;
            });

            return function readChunk() {
              return _ref2.apply(this, arguments);
            };
          })();
          while (yield readChunk()) {
            // Okay to send previous iteration.
            // (For the (n > 1)th iteration--this is handled by the dummy
            // so we don't have to check.)
            waitNextIteration.push(true);
            // Wait for next iteration.
            waitNextIteration = new Box();
            // Send previous chunk.
            yield _this2.sendChunkToIterator(previousChunk, waitNextIteration);
            // Okay to send previous chunk to iterator.
            // Okay to write next chunk.
            streamWaitNext.push(true);

            previousChunk = chunk;
          }
          // Send remaining chunk.
          waitNextIteration.push(true);
          _this2.sendChunkToIterator(chunk); // Nothing to wait for, let's go!
        }

        // End iterator
        waitNextIteration.push(true);
        _this2.sendChunkToIterator(EOS); // Also nothing to wait for here.
      } catch (err) {
        _this2.emit('error', err);
      }
    })();
  }

  // Wait for the iterator to receive a chunk.
  sendChunkToIterator(chunk, waitNext = Promise.resolve()) {
    return new Promise(resolve => {
      this.next = () => {
        this.next = () => {
          throw new Error("Must wait for the current iteration value to resolve.");
        };
        let done = chunk === EOS;
        let v = done ? undefined : chunk;
        return {
          done,
          value: _asyncToGenerator(function* () {
            // Signal that the chunk has been received.
            resolve();
            // Wait for the next chunk to be sent.
            yield waitNext;
            // Return the value already sent.
            return v;
          })()
        };
      };
      // Signal that chunks can be read from the iterator.
      // Rely on Promise#resolve reentry prevention to ensure this only happens
      // once. :)
      this.waitFirstChunk.push(true);
    });
  }

  // Read a chunk from the stream.
  readChunkFromStream(waitNext = Promise.resolve()) {
    return new Promise((resolve, reject) => {
      let h1;
      // Handler called when the promise is resolved or rejected.
      let precomplete = () => {
        // Remove event listeners.
        this.removeListener('error', h1);
      };
      // enc is significant here (argument counting)
      // eslint-disable-next-line no-unused-vars
      this.writeAsync = (() => {
        var _ref4 = _asyncToGenerator(function* (chunk, enc) {
          precomplete();
          // Receive a chunk from the stream.
          resolve(chunk);
          // Wait until the chunk has been consumed.
          yield waitNext;
        });

        return function (_x, _x2) {
          return _ref4.apply(this, arguments);
        };
      })();
      // Synchronous method called when the stream has been ended.
      this.endRequest = () => {
        precomplete();
        // Represent the end of the stream with a symbol EOS.
        // Resolve this special symbol.
        resolve(EOS);
      };
      h1 = () => {
        precomplete();
        // Handle stream errors with a promise rejection here.
        reject();
      };
      this.once('error', h1);
    });
  }

  _write(chunk, enc, cb) {
    nodeify(this.writeAsync(chunk, enc), cb);
  }

  end(chunk, enc, cb) {
    super.end(chunk, enc, err => {
      if (err) {
        cb(err);
        return;
      }
      nodeify(this.endRequest(), cb);
    });
  }

  // Support the iterable interface for convenience.
  [Symbol.iterator]() {
    return this;
  }

};